# -*- coding: utf-8 -*-
import itertools
import json
import logging
import os
import random
import socket
import string
import time
import urlparse

from cronq.config import RABBITMQ_URL
from haigha.connections.rabbit_connection import RabbitConnection
from haigha.message import Message
from cronq.rabbit_connection import parse_url

logger = logging.getLogger(__name__)


def create_host_factory(hosts):
    random.shuffle(hosts)
    hosts_itr = itertools.cycle(hosts)
    return hosts_itr.next


class QueueConnection(object):

    """A wrapper around an AMQP connection for ease of publishing

    Simple instantiation:

        RABBITMQ_URL = os.getenv('RABBITMQ_URL', 'amqp://guest@localhost')
        queueconnection = QueueConnection(RABBITMQ_URL)

    Publishing requires knowing the exchange, routing_key, (headers) and
    the string body:

        # We don't really use headers
        # look up AMQP stuff if you are interested
        queueconnection.publish(
            exchange='exchange',
            routing_key='routing_key',
            headers={},
            body='String Body'
        )

    Usually we want to publish JSON so there is a method that will serialize
    a dict for you:

        # We don't really use headers
        # look up AMQP stuff if you are interested
        queueconnection.publish_json(
            exchange='exchange',
            routing_key='routing_key',
            headers={},
            body={'key': 'value'}
        )

    All of these are asynchronous publishes to the server, if you want to make
    sure that these at least make it to the server you can turn on publisher
    confirmations. Sync calls take longer so they aren't on by default. Turning
    them on is done by passing `confirm=True` to the constructor.

        RABBITMQ_URL = os.getenv('RABBITMQ_URL', 'amqp://guest@localhost')
        queueconnection = QueueConnection(RABBITMQ_URL, confirm=True)

    Now publishes will return a bool of whether the publish succeeded.

        # We don't really use headers
        # look up AMQP stuff if you are interested
        succeeded = queueconnection.publish(
            exchange='exchange',
            routing_key='routing_key',
            headers={},
            body='String Body'
        )

    Without `confirm=True` the return value of publish will only indicate if
    the message was written to a connection successfully.

    """

    def __init__(self, url=None, confirm=False, **kwargs):
        if url is None:
            url = RABBITMQ_URL
        hosts, user, password, vhost, port, heartbeat = parse_url(RABBITMQ_URL)

        if heartbeat is None:
            heartbeat = kwargs.get('heartbeat', None)

        self._connection_hosts = hosts
        self._connection_user = user
        self._connection_password = password
        self._connection_path = vhost
        self._connection_port = port
        self._connection_heartbeat = heartbeat
        self._connection_name = self._generate_connection_name()

        self._connection_params = urlparse.urlparse(url)
        self._connect_attempt_delay = 0.1

        self._get_next_host = create_host_factory(self._connection_hosts)

        self._confirm = confirm
        self._logger = logger

        self._create_connection()

        self._acked = False
        self._last_confirmed_message = None

    def _create_connection(self):
        "Tries to create a connection, returns True on success"
        self._connection = None
        self._last_confirmed_message = None
        host = self._get_next_host()
        self._logger.debug('Trying to connect to {}'.format(host))
        try:
            self._connection = RabbitConnection(
                host=host,
                port=self._connection_port,
                user=self._connection_user,
                password=self._connection_password,
                vhost=self._connection_path,
                close_cb=self._close_cb,
                heartbeat=self._connection_heartbeat,
                client_properties={
                    'connection_name': self._connection_name,
                },
            )
        except socket.error as exc:
            self._logger.error('Error connecting to rabbitmq {}'.format(exc))
            return False
        self._channel = self._connection.channel()
        if self._confirm:
            self._channel.confirm.select(nowait=False)
            self._channel.basic.set_ack_listener(self._ack)
        self._logger.debug('Connected to {}'.format(host))
        return True

    def _generate_connection_name(self):
        random_generator = random.SystemRandom()
        random_string = ''.join([random_generator.choice(string.ascii_lowercase)
                                 for i in xrange(10)])
        return '{0}-{1}-{2}'.format(
            socket.gethostname(),
            os.getpid(),
            random_string,
        )

    def _try_to_connect(self, attempts=3):
        """Try to connect handling retries"""
        for _ in range(attempts):
            if self.is_connected():
                return True
            else:
                self._logger.debug("attempt to create connection")
                self._create_connection()
                time.sleep(self._connect_attempt_delay)

    def _ack(self, message_id):
        self._last_confirmed_message = message_id

    def is_connected(self):
        if self._connection is None or self._channel is None:
            return False
        return True

    def _close_cb(self):
        if self._connection is None:
            reason = 'unknown'
        else:
            reason = self._connection.close_info['reply_text']
        self._logger.info('Disconnected because: {}'.format(reason))
        self._connection = None
        self._channel = None

    def publish(self,
                exchange,
                routing_key,
                headers,
                body,
                connect_attempts=3):
        """Publish a messages to AMQP

        Returns a bool about the success of the publish. If `confirm=True` True
        means it reached the AMQP server. If `confirm=False` it means that it
        was able to be written to a connection but makes no guarantee about the
        message making it to the server.

        """
        if not self.is_connected():
            self._try_to_connect(attempts=connect_attempts)

        if self._connection is None or self._channel is None:
            self._logger.error('Tried to publish without an AMQP connection')
            return False

        msg_number = self._channel.basic.publish(
            Message(body, application_headers=headers),
            exchange=exchange,
            routing_key=routing_key
        )

        if self._confirm:
            if self.is_connected():
                self._connection.read_frames()
                return self._last_confirmed_message == msg_number
            else:
                return False

        return True

    def publish_json(self, exchange, routing_key, headers, body):
        data = json.dumps(body)
        return self.publish(exchange, routing_key, headers, data)

    def close(self):
        self._connection.close()


class Publisher(object):

    def __init__(self):
        self.queue_connection = QueueConnection(RABBITMQ_URL, confirm=True)

    def publish(self, routing_key, job, run_id):
        cmd = {
            'run_id': str(run_id),
            'job_id': job['id'],
            'cmd': job['command'],
            'name': job['name'],
        }
        logger.debug(cmd)
        return self.queue_connection.publish_json("cronq", routing_key, {}, cmd)



# -*- coding: utf-8 -*-
import datetime
import fcntl
import json
import logging
import logging.handlers
import os
import random
import re
import socket
import subprocess
import sys
import time

from cronq.config import LOG_PATH
from cronq.config import QUEUE
from cronq.rabbit_connection import CronqConsumer
from cronq.utils import unicodedammit

from haigha.message import Message

FILENAME_REGEX = re.compile('[\W_]+', re.UNICODE)


class CronqRunner(CronqConsumer):

    """handles reconnects"""

    def publish_result(self, body):
        headers = {
            'x-send-datetime': str(datetime.datetime.utcnow()),
            'x-host': str(socket.getfqdn()),
        }
        body.update(headers)
        msg = Message(json.dumps(body))
        self.publish(msg, 'cronq', 'cronq_results')

    def valid_job(self, data):
        valid = True
        for key in ['cmd', 'job_id', 'run_id']:
            if not data.get(key, None):
                self.logger.debug('Missing {0}'.format(key))
                valid = False
        return valid

    def run_job(self, data):
        start = time.time()
        process = None

        cmd = data.get('cmd')
        job_id = data.get('job_id')
        run_id = data.get('run_id')

        self.log_message(job_id, run_id, "Starting {}".format(cmd))

        # communicate start
        self.publish_result({
            'job_id': data.get('job_id'),
            'run_id': data.get('run_id'),
            'start_time': str(datetime.datetime.utcnow()),
            'type': 'starting',
        })

        try:
            process = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
        except OSError:
            # log exception with stack trace
            self.logger.exception('[cronq_job_id:{0}] [cronq_run_id:{1}] Failed job'.format(
                data.get('job_id'), data.get('run_id')
            ))
            end = time.time()

            # communicate failure
            self.publish_result({
                'job_id': job_id,
                'run_id': run_id,
                'run_time': end - start,
                'type': 'failed',
            })

            return False

        # process job output
        fd = process.stdout.fileno()
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

        self.log_message(job_id, run_id, "Waiting")

        splits = FILENAME_REGEX.split(data.get('name', 'UNKNOWN'))
        if len(splits) > 1:
            logfile = '-'.join(splits)
        filename = '{0}/{1}.log'.format(LOG_PATH, logfile.strip('-'))

        handler = logging.handlers.WatchedFileHandler(filename)
        log_to_stdout = bool(os.getenv('CRONQ_RUNNER_LOG_TO_STDOUT', False))

        while True:
            try:
                nextline = process.stdout.readline()
            except IOError:
                nextline = ''

            if nextline == '' and process.poll() is not None:
                break

            if nextline == '':
                time.sleep(0.1 * random.random())
                continue

            try:
                message = nextline.rstrip()
                message = unicodedammit(message)
            except (TypeError, ValueError):
                continue

            if message:
                for m in message.splitlines():
                    log_record = logging.makeLogRecord({
                        'msg': message,
                    })
                    handler.emit(log_record)
                    if log_to_stdout:
                        self.log_message(
                            job_id, run_id, log_record.getMessage())

            time.sleep(0.00001)
            sys.stdout.flush()

        handler.close()

        # communicate finished
        end = time.time()
        self.publish_result({
            'job_id': job_id,
            'run_id': run_id,
            'return_code': process.returncode,
            'run_time': end - start,
            'type': 'finished',
        })

        self.log_message(job_id, run_id, "[cronq_exit_code:{}] Done".format(process.returncode))
        return True

    def run_something(self, msg):
        make_directory(LOG_PATH)

        data = json.loads(str(msg.body))
        if not self.valid_job(data):
            return self.reject(requeue=False)

        try:
            success = self.run_job(data)
        except KeyboardInterrupt:
            raise
        except:
            self.logger.exception("unhandled exception in job")
            success = False

        if not success:
            self.logger.warning("rejecting message")
            return self.reject(msg, requeue=False)
        else:
            self.logger.info("acking message")
            return self.ack(msg)


def setup():
    runner = CronqRunner()

    max_failures = 1000
    while max_failures > 0:
        runner.connect()
        runner.consume(queue=QUEUE)
        max_failures -= 1

    runner.logger.warning("Too many errors, exiting")
    sys.exit(1)


def make_directory(directory):
    try:
        os.mkdir(directory)
    except OSError:
        pass


def main():
    setup()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import json
import logging
from uuid import UUID
import sys

from cronq.backends.mysql import Storage
from cronq.rabbit_connection import CronqConsumer
from dateutil.parser import parse

logger = logging.getLogger(__name__)


class CronqAggregator(CronqConsumer):

    def __init__(self):
        super(CronqAggregator, self).__init__()
        self.storage = Storage()

    def run_something(self, msg):
        data = json.loads(str(msg.body))

        job_id = data['job_id']
        run_id = UUID(hex=data['run_id'])

        self.log_message(job_id, run_id,
                         "Write job result to database {}".format(str(msg.body)))

        self.storage.add_event(
            job_id,
            parse(data.get('x-send-datetime')),
            run_id.hex,
            data.get('type'),
            data.get('x-host'),
            data.get('return_code'),
        )

        self.log_message(job_id, run_id, "Attempting to update status on job")

        self.storage.update_job_status(
            run_id,
            data.get('job_id'),
            parse(data.get('x-send-datetime')),
            data.get('type'),
            data.get('return_code', None))

        self.log_message(job_id, run_id, "Acking message on job")
        self.ack(msg)


def setup():
    runner = CronqAggregator()

    max_failures = 10
    while max_failures > 0:
        runner.connect()
        runner.consume(queue="cronq_results")
        max_failures -= 1

    runner.logger.warning("Too many errors, exiting")
    sys.exit(1)


def main():
    setup()

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import datetime


def split_command(string):
    commands = string.strip().split(';')
    ret = []
    for command in commands:
        ret.extend(command.strip().split(' && '))
    return ret


def task_status(first, last=None):
    if last is None:
        return first['status']

    if last['status'] == 'finished':
        if int(last['return_code']) == 0:
            return 'succeeded'
        return 'failed'
    return last['status']


def took(first_time, last_time):
    if type(first_time) is not datetime.datetime:
        return ""

    if type(last_time) is not datetime.datetime:
        return ""

    elapsed_time = last_time - first_time
    return int(elapsed_time.total_seconds())


def unicodedammit(x):
    """encode as string, decode as unicode bytes"""
    try:
        decoded = decode_string(x, ignore=False)
        return decoded
    except Exception:
        # decide with ignore
        decoded = decode_string(x, ignore=True)
        return decoded
    return decoded


def decode_string(newstr, ignore=False):
    """attempt to decode the string with three different options
        ascii
        utf8mb4
        utf8
        latin1
    If none of these works raise an error
    """

    # this is already unicode. return
    if type(newstr) is unicode:
        return newstr

    encodings = ['ascii', 'utf8mb4', 'utf8', 'latin1']
    for encoding in encodings:
        try:
            if ignore:
                dec_str = newstr.decode(encoding, 'ignore')
            else:
                dec_str = newstr.decode(encoding)
            return dec_str
        except:
            pass

    raise Exception("can't decode this string at all")



# -*- coding: utf-8 -*-
import logging
import os
import random
import socket
import string
import time
import urlparse

from cronq.config import RABBITMQ_URL

from haigha.connections.rabbit_connection import RabbitConnection

# cronq jobs can be very long, override heartbeat settings
# to keep rabbitmq connections alive (or there might be problems)
CONSUMER_HEARTBEAT = 50000


class NullHandler(logging.Handler):

    def emit(self, record):
        pass

logger = logging.getLogger(__name__)


def generate_random_string(length):
    """generates  a random alphanumeric string of length `strlen`"""
    random_generator = random.SystemRandom()
    return ''.join([random_generator.choice(string.ascii_lowercase) for i in xrange(length)])


def parse_heartbeat(query):
    default_heartbeat = None
    heartbeat = default_heartbeat
    if query:
        qs = urlparse.parse_qs(query)
        heartbeat = qs.get('heartbeat', default_heartbeat)
    else:
        logger.debug('No heartbeat specified, using broker defaults')

    if isinstance(heartbeat, (list, tuple)):
        if len(heartbeat) == 0:
            logger.warning('No heartbeat value set, using default')
            heartbeat = default_heartbeat
        elif len(heartbeat) == 1:
            heartbeat = heartbeat[0]
        else:
            logger.warning('Multiple heartbeat values set, using broker default: {0}'.format(
                heartbeat
            ))
            heartbeat = default_heartbeat

    if type(heartbeat) == str and heartbeat.lower() == 'none':
        return None

    if heartbeat is None:
        return heartbeat

    try:
        heartbeat = int(heartbeat)
    except ValueError:
        logger.warning('Unable to cast heartbeat to int, using broker default: {0}'.format(
            heartbeat
        ))
        heartbeat = default_heartbeat

    return heartbeat


def parse_url(rabbitmq_url):
    """returns tuple containing
    HOSTS, USER, PASSWORD, VHOST
    """
    hosts = user = password = vhost = None
    port = 5672

    cp = urlparse.urlparse(rabbitmq_url)
    hosts_string = cp.hostname
    hosts = hosts_string.split(",")
    if cp.port:
        port = int(cp.port)
    user = cp.username
    password = cp.password
    vhost = cp.path
    query = cp.query

    port = 5672
    if cp.port:
        port = cp.port

    # workaround for bug in 12.04
    if '?' in vhost and query == '':
        vhost, query = vhost.split('?', 1)

    heartbeat = parse_heartbeat(query)
    return (hosts, user, password, vhost, port, heartbeat)


def plain_rabbit_connection_to_hosts(hosts, **kwargs):
    for host in hosts:
        logger.info('Trying to connect to host: {0}'.format(host))
        try:
            conn = RabbitConnection(host=host, **kwargs)
            logger.info("...success")
            return conn
        except socket.error:
            logger.info('Error connecting to {0}'.format(host))
    logger.error('Could not connect to any hosts')


def connect(**kwargs):
    hosts, user, password, vhost, port, heartbeat = parse_url(RABBITMQ_URL)

    logger.info('Hosts are: {0}'.format(hosts))
    rabbit_logger = logging.getLogger('dispatcher.haigha')

    random_string = generate_random_string(10)

    connection_name = '{0}-{1}-{2}'.format(
        socket.gethostname(),
        os.getpid(),
        random_string,
    )

    heartbeat = heartbeat or None

    # override connection string sometimes
    if kwargs.get('heartbeat'):
        heartbeat = kwargs.get('heartbeat')

    close_cb = kwargs.get('close_cb')

    conn = plain_rabbit_connection_to_hosts(
        hosts,
        port=port,
        user=user,
        password=password,
        vhost=vhost,
        heartbeat=heartbeat,
        logger=rabbit_logger,
        close_cb=close_cb,
        client_properties={
            'connection_name': connection_name
        }
    )

    return conn


class CronqConsumer(object):

    """handles reconnects"""

    def __init__(self):
        self._connection = None
        self._channel = None

        # logging
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(NullHandler())

    def _create_connection(self):
        self._connection = connect(
            close_cb=self._connection_closed_cb,
            heartbeat=CONSUMER_HEARTBEAT)
        self._channel = self._connection.channel()

    def connect(self):
        return self._create_connection()

    @property
    def connection(self):
        if self._connection:
            return self._connection

    @property
    def channel(self):
        if self._channel:
            return self._channel

    def is_connected(self):
        return self.connection and self.channel

    def make_connected(self, attempts=3):
        for _ in range(attempts):
            if self.is_connected():
                return True
            else:
                self.logger.debug("attempt to create connection")
                self._create_connection()
                time.sleep(.1)
        return False

    @staticmethod
    def tag_from_msg(msg):
        tag = msg.delivery_info['delivery_tag']
        return tag

    def ack(self, msg):
        tag = self.tag_from_msg(msg)
        if self.is_connected():
            self.channel.basic.ack(tag)

    def reject(self, msg, requeue=True):
        tag = self.tag_from_msg(msg)
        if self.is_connected():
            self.channel.basic.reject(tag, requeue=requeue)

    def publish(self, msg, exchange, routing_key):
        self.channel.basic.publish(msg, exchange, routing_key)

    def log_message(self, job_id, run_id, message, lvl=logging.INFO):
        msg = "[cronq_job_id:{0}] [cronq_run_id:{1}] {2}".format(
            job_id,
            run_id,
            message
        )
        self.logger.log(lvl, msg)

    def run_something(self, msg):
        raise NotImplementedError()

    def consume(self, queue):
        self.make_connected()

        def runner(msg):
            return self.run_something(msg)

        self.channel.basic.qos(prefetch_count=1)
        self.channel.basic.consume(
            queue=queue,
            consumer=runner,
            no_ack=False
        )

        while self.is_connected():
            self.connection.read_frames()

        return False

    def _connection_closed_cb(self):
        if self._connection is None:
            reason = 'unknown'
        else:
            reason = self._connection.close_info['reply_text']
        self.logger.info('Disconnected because: {}'.format(reason))
        self._connection = None
        self._channel = None
